/******************************************************************************
* BIM M 13x Serial Communication
*
* File: application.c
*
* Date: 2009/03/16
*
* Version: 002
*
* Author: Schiekofer
******************************************************************************/
#include "BIM_M13x.h"
#include "BIM_M_13x_Serial_LowLevel.h"
#include "BIM_M_13x_Serial_GrpAddr.h"
#include "application.h"

#define APCI_GROUPVALUE_WRITE     0x0080
#define APCI_GROUPVALUE_READ      0x0000
#define APCI_GROUPVALUE_RESPONSE  0x0040

MESSAGEQUEUE UserQueue;
MESSAGE* pMsg;
TIMER_TAB UserTimerTab;

void AppInit(void)
{
    // clear EXPRAM
    for (USHORT x = 0xF000; x < 0xF400; x++)
    {
        *(BYTE*)x = 0x00;
    }

    // init timers    
    U._TmInit(NUM_OF_TIMERS);
    // init serial communication modules
    SerialLowLevel_Init();
    SerialGrpAddr_Init();
    
    // get all messages from linklayer in 'UserQueue1'
    U._MsgSwitchQueue(&U._pDynQueues->NetworkLayer, &UserQueue);
    
    // after initialization a taskswitch must be done because
    // the task switch is disabled at cstartup
    FORCE_TASKSWITCH();
}

//------------------------------------------------------------------------
//  This function is called if there was an
//  unload of the application program
//------------------------------------------------------------------------
void unload()
{
    MESSAGE* pMsg;
    
    // reset message queue structure
    U._MsgResetDynQueues();
    
    // get all messages that are still in UserQueue
    pMsg = U._MsgGet(&UserQueue);
    while (pMsg != NULL)
    {
        // post it back to system
        U._MsgPost(pMsg, &U._pStaticQueues->NetworkLayer);
        pMsg = U._MsgGet(&UserQueue);        
    }
}

void Application_Task(void)
{
    USHORT apci;
    
    // cyclically call the serial 'low level' task
    SerialLowLevel_Task();
    
    // check if there is a message in UserQueue
    pMsg = U._MsgGet(&UserQueue);
    if (pMsg != NULL)
    {
        // get destination address of the message
        WORD destaddr = Word(pMsg->Data[3], pMsg->Data[4]);
        // process message only if group addressed and no broadcast
        if ( ((pMsg->Data[5] & 0x80) == 0x80) && (destaddr != 0x0000) )
        {
            BYTE length;
            BYTE datastart;

            // get apci
            apci = (pMsg->Data[6] << 8) | pMsg->Data[7];
            apci = apci & 0x03C0;                
            
            // get length
            length = pMsg->Data[5] & 0x0F;
            // check if payload is in byte 7 or 8
            if (length == 1)
            {
                pMsg->Data[7] = pMsg->Data[7] & 0x3F;
                datastart = 7;
            }
            else
            {
                datastart = 8;
            }
            length--;

            // check if message is an indication or a confirm
            if ((pMsg->ServiceCode & (LAYER_MASK + TYPE_MASK)) == (LL + IND))
            {
                // send the corresponding indication 
                switch (apci)
                {
                case APCI_GROUPVALUE_WRITE:
                    SGA_WriteIndication(destaddr, &pMsg->Data[datastart], length);                    
                    break;
                case APCI_GROUPVALUE_READ:
                    SGA_ReadIndication(destaddr);
                    break;
                case APCI_GROUPVALUE_RESPONSE:
                    SGA_ResponseIndication(destaddr, &pMsg->Data[datastart], length);
                    break;
                default:
                    break;
                }
            }
            else if ((pMsg->ServiceCode & (LAYER_MASK + TYPE_MASK)) == (LL + CON))
            {
                // send the corresponding confirm
                switch (apci)
                {
                case APCI_GROUPVALUE_WRITE:
                    SGA_WriteConfirm(destaddr, &pMsg->Data[datastart], length);
                    break;
                case APCI_GROUPVALUE_READ:
                    SGA_ReadConfirm(destaddr);
                    break;
                case APCI_GROUPVALUE_RESPONSE:
                    SGA_ResponseConfirm(destaddr, &pMsg->Data[datastart], length);
                    break;
                default:
                    break;
                }
            }            
            // always discard message after processing
            U._MsgDiscard(pMsg);
        }
        else
        {
            // message is not group addressed or a broadcast -> post it back to system
            U._MsgPost(pMsg, &U._pStaticQueues->NetworkLayer);
        }
    }    
}
