/******************************************************************************
* BIM M 13x Serial Communication
*
* File: BIM_M_13x_SerialCom_LowLevel.cs
*
* Date: 2009/03/16
*
* Version: 002
*
* Author: Schiekofer
******************************************************************************/
using System;
using System.Collections.Generic;
using System.Text;
using System.IO;
using System.IO.Ports;
using System.Threading;

namespace BIM_M_13x_SerialCom
{
    public abstract class BIM_M_13x_SerialCom_LowLevel
    {
        #region PUBLIC_ENUMS
        public enum BIM_M_13x_SerialCom_LowLevel_System
        {
            windows = 0,
            linux = 1,
        }
        #endregion
        #region PUBLIC_EVENTS
        public delegate void OnReceiveError();
        public event BIM_M_13x_SerialCom_LowLevel.OnReceiveError ErrorHandler;
        #endregion
        #region PUBLIC_MEMBERS
        public BIM_M_13x_SerialCom_LowLevel()
        {
            // configure serial port
            m_sp = new SerialPort();                
            m_sp.BaudRate = 19200;
            m_sp.DataBits = 8;
            m_sp.StopBits = StopBits.One;
            m_sp.Parity = Parity.Even;
            m_sp.Handshake = Handshake.None;
            m_sp.ReadTimeout = RECEIVE_TIMEOUT;

            // create new message queues
            m_rqueue = new Queue<byte[]>();
            m_tqueue = new Queue<byte[]>();

            // create new thread controlling instances
            m_mut_trm = new Mutex();
            m_mut_rcv = new Mutex();
            m_are_trm = new AutoResetEvent(false);
            m_are_rcv = new AutoResetEvent(false);
            m_are_timeout_start = new AutoResetEvent(false);
            m_are_timeout_stop = new AutoResetEvent(false);
            m_are_timeout_is_stopped = new AutoResetEvent(false);

            // all threads check this variable for exit
            m_ExitNow = true;            
        }
        public void Connect(int ComPort, BIM_M_13x_SerialCom_LowLevel_System sys)
        {
            // first disconnect (could always be called)
            Disconnect();

            // get choosen system
            switch (sys)
            {
                case BIM_M_13x_SerialCom_LowLevel_System.windows:
                    m_sp.PortName = "COM" + ComPort.ToString();
                    break;
                case BIM_M_13x_SerialCom_LowLevel_System.linux:
                    m_sp.PortName = "/dev/ttyS" + ComPort.ToString();
                    break;
                default:
                    throw new Exception("Unknown system!");
            }                        

            // init ack indicators
            m_send_ack_nack = NOTHING;
            m_ack_nack_received = NOTHING;
            
            // init flow control variables
            m_tm_expired = false;
            m_start_receiving = false;
            m_reset = BIM_M_13x_SerialCom_LowLevel_ResetState.idle;

            // init receive write pointer
            m_writepointer = 0;

            // init sequence counters
            m_rcvSeqCounter = 0;
            m_trmSeqCounter = 0;

            // clear message queues
            m_rqueue.Clear();
            m_tqueue.Clear();

            // indicate all threads that running is allowed but nothing to do
            m_ExitNow = false;
            m_are_rcv.Reset();
            m_are_trm.Reset();
            m_are_timeout_start.Reset();
            m_are_timeout_stop.Reset();
            m_trmstate = BIM_M_13x_SerialCom_LowLevel_TRM_States.idle;

            // create new threads
            m_transmitthread = new Thread(new ThreadStart(TransmitThread));
            m_receiverthread = new Thread(new ThreadStart(ReceiverThread));
            m_receiverworkingthread = new Thread(new ThreadStart(ReceiverWorkingThread));
            m_timeoutthread = new Thread(new ThreadStart(TimeoutThread));

            // open serial port
            m_sp.Open();

            // start all threads
            m_transmitthread.Start();
            m_receiverthread.Start();
            m_receiverworkingthread.Start();
            m_timeoutthread.Start();
        }
        public void Disconnect()
        {
            if (m_ExitNow == false)
            {
                // indicate all threads that they must run and must exit
                m_ExitNow = true;
                m_are_rcv.Set();
                m_are_trm.Set();
                m_are_timeout_start.Set();
                m_are_timeout_stop.Set();

                // Disconnect() could be called even if no thread instance was 
                // created, therefore check if thread is not null and thread is alive
                // if it is alive, wait for exit
                if (m_transmitthread != null)
                {
                    if (m_transmitthread.Join(250) == false)
                    {
                        m_transmitthread.Abort();
                    }
                }
                if (m_receiverworkingthread != null)
                {
                    if (m_receiverworkingthread.Join(250) == false)
                    {
                        m_receiverworkingthread.Abort();
                    }
                }
                if (m_receiverthread != null)
                {
                    if (m_receiverthread.Join(250) == false)
                    {
                        m_receiverthread.Abort();
                    }
                }
                if (m_timeoutthread != null)
                {
                    if (m_timeoutthread.Join(250) == false)
                    {
                        m_timeoutthread.Abort();
                    }
                }
            }

            // close serial port if it is open
            if (m_sp.IsOpen == true)
            {
                m_sp.Close();
            }
        }
        #endregion
        #region PROTECTED_MEMBERS
        // this function must be overwritten in derived class
        protected abstract void OnReceiveMsg(byte[] msg);
        protected bool SendMsg(byte[] msg)
        {
            bool returnval;
            byte[] ba = new byte[msg.Length + 3];
            // copy msg and add space for Length, SequenceCounter and Checksum
            Array.Copy(msg, 0, ba, 2, msg.Length);

            // check if threads should run
            if (m_ExitNow == false)
            {
                // wait for exclusive accesss to transmit queue
                m_mut_trm.WaitOne();
                // check if transmit queue is not too full
                if (m_tqueue.Count < MAX_TRANSMIT_MESSAGES)
                {
                    // put message into transmit queue
                    m_tqueue.Enqueue(ba);
                    // indicate that there is a message in the queue
                    m_are_trm.Set();
                    // set return value;
                    returnval = true;
                }
                else 
                {
                    // too many messages in the queue
                    returnval = false;
                }
                // allow other threads access to the transmit queue
                m_mut_trm.ReleaseMutex();
            }
            else
            {
                returnval = false;
            }
            return returnval;
        }
        #endregion
        #region PRIVATE_ENUMS
        private enum BIM_M_13x_SerialCom_LowLevel_TRM_States
        {
            idle = 0,
            wait_for_ack = 1,
        }
        private enum BIM_M_13x_SerialCom_LowLevel_ResetState
        {
            idle = 0,
            reset_req = 1,
            reset_conf = 2,
        }
        #endregion
        #region PRIVATE_MEMBERS
        private void ReceiverThread()
        {
            // this thread is running until 'm_ExitNow == true'
            while (true)
            {
                while (m_ExitNow == false)
                {
                    // read from the serial port
                    try
                    {
                        if (m_sp.IsOpen)
                        {
                            m_writepointer += m_sp.Read(m_rbuff, m_writepointer, m_rbuff.Length - m_writepointer);
                        }
                        else
                        {
                            CommunicationError();
                            return;
                        }
                    }
                    catch (TimeoutException)
                    {                        
                        // timeout occurs after RECEIVE_TIMEOUT
                        // and starts a check for a received message or ACK/NACK
                        if (m_start_receiving == true)
                        {   // if this is true, the receiving of an telegram
                            // was not finished in RECEIVE_TIMEOUT time
                            // --> send NACK and clear receive buffer
                            m_start_receiving = false;
                            m_writepointer = 0;
                            m_send_ack_nack = NACK;
                            m_are_trm.Set();
                            continue;
                        }
                    }
                    // do a break here and check for a received message
                    break;
                }

                // check for exit
                if (m_ExitNow == true)
                {                        
                    return;
                }

                // first clear buffer from ACK/NACK bytes (normaly only 1)
                while (((m_rbuff[0] & ACK_MASK) > 0) && (m_writepointer > 0))
                {
                    // set ACK/NACK value
                    m_ack_nack_received = m_rbuff[0];
                    // delete byte in receive buffer
                    Array.Copy(m_rbuff, 1, m_rbuff, 0, (m_writepointer - 1));
                    m_writepointer--;
                }

                // if there are more byte, check for a received message
                if (m_writepointer > 0)
                {                    
                    CheckForMessage();
                }
            }
        }
        private void ReceiverWorkingThread()
        {
            byte[] ba;
            // this thread is running until 'm_ExitNow == true'
            while (true)
            {
                // wait for exclusive accesss to the receive queue
                m_mut_rcv.WaitOne();
                // wait for a messages in the receive queue
                while ((m_ExitNow == false) && (m_rqueue.Count == 0))
                {
                    // allow other threads to access the receive queue
                    m_mut_rcv.ReleaseMutex();
                    // wait for a signal that indicates that there a messages
                    // in the queue
                    m_are_rcv.WaitOne();
                    // wait for exclusive accesss to the receive queue
                    m_mut_rcv.WaitOne();
                }
                // here 'm_rqueue.Count > 0'
                // first check again for exit
                if (m_ExitNow == true)
                {                    
                    m_mut_rcv.ReleaseMutex();                        
                    return;
                }
                // get the message from the queue
                ba = (byte[])m_rqueue.Dequeue();
                // allow other threads to access the receive queue
                m_mut_rcv.ReleaseMutex();
                // call the receive handling function
                OnReceiveMsg(ba);
            }
        }
        private void TransmitThread()
        {            
            int queuecount;
            byte[] ba;

            // this thread is running until 'm_ExitNow == true'
            while (true)
            {
                // wait for a signal that indicates that there is a message to send
                m_are_trm.WaitOne();

                // check for exit
                if (m_ExitNow == true)
                {                        
                    return;
                }
                // first check if an ACK/NACK is to send
                if (m_send_ack_nack != NOTHING)
                {
                    if (WriteByte(m_send_ack_nack) == false)
                    {
                        return;
                    }
                    m_send_ack_nack = NOTHING;
                }
                // check the actual state
                switch (m_trmstate)
                {
                    case BIM_M_13x_SerialCom_LowLevel_TRM_States.idle:
                        // wait for exclusive accesss to the transmit queue
                        m_mut_trm.WaitOne();
                        // get number of messages
                        queuecount = m_tqueue.Count;
                        // allow other threads to access the transmit queue
                        m_mut_trm.ReleaseMutex();
                        // first check for a reset request (send a reset message)
                        if (m_reset != BIM_M_13x_SerialCom_LowLevel_ResetState.idle)
                        {
                            // generate a reset message
                            byte[] reset = {0x04, 0x00, RESET, 0x00};
                            // add checksum to the reset messsage
                            AddChecksum(ref reset);
                            // set reset indicator to confirm
                            m_reset = BIM_M_13x_SerialCom_LowLevel_ResetState.reset_conf;
                            if (m_sp.IsOpen == true)
                            {
                                // send reset message
                                m_sp.Write(reset, 0, reset.Length);
                            }
                            else
                            {
                                CommunicationError();
                                return;
                            }
                            // reset ACK/NACK indicator
                            m_ack_nack_received = NOTHING;
                            // wait for ack
                            m_trmstate = BIM_M_13x_SerialCom_LowLevel_TRM_States.wait_for_ack;
                            // start ack timeout time
                            StartTimeoutTimer(ACK_TIMEOUT);
                        }
                        else if (queuecount > 0)
                        {
                            // if there are messages to send
                            // wait for exclusive accesss to the transmit queue
                            m_mut_trm.WaitOne();
                            // just copy to first message from the queue
                            ba = (byte[])m_tqueue.Peek();
                            // allow other threads to access the transmit queue
                            m_mut_trm.ReleaseMutex();

                            // check if serial port is open
                            if (m_sp.IsOpen)
                            {
                                // add length to the message
                                ba[0] = (byte)(ba.Length & 0x1F);
                                // add actual transmit sequence counter
                                ba[1] = m_trmSeqCounter;
                                // add checksum
                                AddChecksum(ref ba);
                                // reset ACK/NACK indicator
                                m_ack_nack_received = NOTHING;
                                // send message
                                m_sp.Write(ba, 0, ba.Length);
                                // wait for ack
                                m_trmstate = BIM_M_13x_SerialCom_LowLevel_TRM_States.wait_for_ack;
                                // start ack timeout time
                                StartTimeoutTimer(ACK_TIMEOUT);
                            }
                            else
                            {
                                // wait for exclusive accesss to the transmit queue
                                m_mut_trm.WaitOne();
                                // clear all messages in transmit queue
                                m_tqueue.Clear();
                                // allow other threads to access the transmit queue
                                m_mut_trm.ReleaseMutex();
                                // go to idle
                                m_trmstate = BIM_M_13x_SerialCom_LowLevel_TRM_States.idle;
                                // call error event
                                CommunicationError();
                                return;
                            }
                        }
                        break;
                    case BIM_M_13x_SerialCom_LowLevel_TRM_States.wait_for_ack:
                        // first check if ACK/NACK was received
                        if (m_ack_nack_received != NOTHING)
                        {
                            // stop ack timeout timer
                            StopTimeoutTimer();

                            // if an ACK was received
                            if (m_ack_nack_received == ACK)
                            {                                
                                if (m_reset == BIM_M_13x_SerialCom_LowLevel_ResetState.reset_conf)
                                {
                                    // if message was a reset message
                                    // set reset indicator to idle
                                    m_reset = BIM_M_13x_SerialCom_LowLevel_ResetState.idle;
                                    // reset receive sequence counter
                                    m_rcvSeqCounter = 0;
                                }
                                else
                                {
                                    // message was a normal message
                                    // wait for exclusive accesss to the transmit queue
                                    m_mut_trm.WaitOne();
                                    // now take the message out of the transmit queue
                                    m_tqueue.Dequeue();
                                    // allow other threads to access the transmit queue                         
                                    m_mut_trm.ReleaseMutex();
                                }
                                // increment transmit sequence counter
                                m_trmSeqCounter++;
                            }
                            // reset ACK/NACK indicator
                            m_ack_nack_received = NOTHING;
                            // go to idle
                            m_trmstate = BIM_M_13x_SerialCom_LowLevel_TRM_States.idle;
                            // trigger a new run in idle mode to check if there is something more to do
                            m_are_trm.Set();
                        }
                        else if (m_tm_expired == true)
                        {
                            // timeout is occured
                            // go to idle
                            m_trmstate = BIM_M_13x_SerialCom_LowLevel_TRM_States.idle;
                            // trigger a new run -> message is send again
                            m_are_trm.Set();
                        }
                        break;
                }
            }
        }
        private void TimeoutThread()
        {
            while (m_ExitNow == false)
            {
                m_are_timeout_is_stopped.Set();
                m_are_timeout_start.WaitOne();
                if (m_ExitNow == true)
                {
                    return;
                }
                m_are_timeout_is_stopped.Reset();
                if (m_are_timeout_stop.WaitOne(m_timeout_interval, false) == false)
                {
                    // set timer expired indicator
                    m_tm_expired = true;
                    // trigger transmit thread
                    m_are_trm.Set();
                }            
            }
            return;
        }
        private void StartTimeoutTimer(int interval)
        {
            m_are_timeout_stop.Set();
            m_are_timeout_is_stopped.WaitOne();
            m_timeout_interval = interval;
            m_tm_expired = false;
            m_are_timeout_stop.Reset();
            m_are_timeout_start.Set();
        }
        private void StopTimeoutTimer()
        {
            m_are_timeout_stop.Set();            
        }
        private void CheckForMessage()
        {            
            byte[] ba;

            // check if too much bytes received
            if (m_writepointer > (MAX_MESSAGE_LENGTH + 1))
            {
                // send NACK
                m_send_ack_nack = NACK;
                // discard all received bytes
                m_sp.DiscardInBuffer();
                // clear receive buffer
                m_writepointer = 0;                                
                return;
            }
            // are there some received bytes
            if (m_writepointer > 0)
            {
                // are there all bytes of the message received
                if (m_writepointer >= m_rbuff[MESSAGE_LENGTH_POS])
                {
                    // check checksum of the message
                    if (CheckChecksum(m_rbuff, 0, m_rbuff[MESSAGE_LENGTH_POS]) == true)
                    {                        
                        // check if a reset was send
                        if (m_rbuff[MESSAGE_CMD_POS] == RESET)
                        {                           
                            // indicate an ACK to send
                            m_send_ack_nack = ACK;                           
                            m_rcvSeqCounter++;
                            // reset transmit sequence counter
                            m_trmSeqCounter = 0;

                            if (m_trmstate == BIM_M_13x_SerialCom_LowLevel_TRM_States.wait_for_ack)
                            {
                                StartTimeoutTimer(ACK_TIMEOUT_SHORT);
                            }
                        }
                        else if (m_rbuff[1] == m_rcvSeqCounter)
                        {
                            // generate a reduced message (without Length, SequCounter, Checksum)
                            ba = new byte[m_rbuff[0] - 3];

                            // copy payload
                            Array.Copy(m_rbuff, 2, ba, 0, ba.Length);

                            // wait for exclusive accesss to the receive queue
                            m_mut_rcv.WaitOne();
                            // put message into the receive queue
                            m_rqueue.Enqueue(ba);
                            // set signal that there is a message in the queue
                            m_are_rcv.Set();
                            // allow other threads to access the receive queue                         
                            m_mut_rcv.ReleaseMutex();
                            // indicate an ACK to send
                            m_send_ack_nack = ACK;
                            // increment receive sequence counter
                            m_rcvSeqCounter++;
                        }
                        else if (m_rbuff[MESSAGE_SEQUCOUNT_POS] == (m_rcvSeqCounter - 1))
                        {
                            // this message is a repeated message, because sequence counter
                            // is one lower than the expected one                            
                            // -> indicate only an ACK to send (no processing of the message!)
                            m_send_ack_nack = ACK;
                        }
                        else
                        {
                            // wrong sequence counter -> indicate a reset to send
                            m_reset = BIM_M_13x_SerialCom_LowLevel_ResetState.reset_req;
                        }
                        // reset the 'start receiving' indicator
                        m_start_receiving = false;
                        // remove the message from the receive buffer
                        m_writepointer = RemoveMsg((m_rbuff[0]), m_writepointer);
                    }
                    else
                    {
                        // wrong checksum -> indicate a NACK to send
                        m_send_ack_nack = NACK;
                        // clear receive buffer
                        m_writepointer = 0;
                    }
                }
                else
                {
                    // some bytes are received but not all message bytes
                    // -> indicate a 'start receiving'
                    // -> message must completely be received
                    //    within the receive timeout of the serial port
                    m_start_receiving = true;
                }
                // trigger transmit thread (for ACK)
                m_are_trm.Set();
            }                                                                                        
        }
        private int RemoveMsg(int start, int end)
        {
            // move all bytes from 'start' to 'end' to position 0 
            Array.Copy(m_rbuff, start, m_rbuff, 0, (end - start));

            // return the first free position after the copied bytes
            return (end - start);            
        }
        private void AddChecksum(ref byte[] msg)
        {         
            // start checksum with 0xA5
            byte check = 0xA5;                        

            // add all bytes modulo 0xFF
            for (int i = 0; i < (msg.Length - 1); i++)
            {
                check = (byte)((check + msg[i]) & 0xFF);
            }
            // enter at last position of the message the calculated checksum value
            msg[msg.Length - 1] = check;
            return;
        }
        private bool CheckChecksum(byte[] ba, int start, int end)
        {
            int i;
            // start checksum with 0xA5
            byte check = 0xA5;

            // add all bytes modulo 0xFF
            for (i = start; i < (end - 1); i++)
            {
                check += ba[i];
            }
            // return whether the last byte of the message is equal to
            // the calculated checksum or not
            return (check == ba[i]) ? true : false;
        }
        private bool WriteByte(byte b)
        {
            // create an array with only one entry
            m_singlebyte[0] = b;
            // check if serial port is open
            if (m_sp.IsOpen)
            {
                // write the single byte array
                m_sp.Write(m_singlebyte, 0, 1);
                return true;
            }
            else
            {
                CommunicationError();
                return false;
            }
        }
        private void CommunicationError()
        {
            // create a new thread
            m_errorthread = new Thread(new ThreadStart(ErrorThread));
            // start the thread
            m_errorthread.Start();
        }
        private void ErrorThread()
        {
            // check if the user of this claas wants to get the error
            if (ErrorHandler != null)
            {
                ErrorHandler();
            }
        }
        #endregion
        #region PRIVATE_VARIABLES
        private byte[] m_rbuff = new byte[RBUFFER_SIZE];
        private byte[] m_singlebyte = new byte[1];
        private byte m_send_ack_nack;
        private byte m_ack_nack_received;
        private byte m_rcvSeqCounter;
        private byte m_trmSeqCounter;
        private bool m_tm_expired;
        private bool m_start_receiving;
        private BIM_M_13x_SerialCom_LowLevel_ResetState m_reset;
        private BIM_M_13x_SerialCom_LowLevel_TRM_States m_trmstate;
        private int m_writepointer;
        private Thread m_transmitthread;
        private Thread m_receiverthread;
        private Thread m_receiverworkingthread;
        private Thread m_errorthread;
        private Thread m_timeoutthread;
        private int m_timeout_interval;
        private bool m_ExitNow;
        private SerialPort m_sp;
        private Queue<byte[]> m_rqueue;
        private Queue<byte[]> m_tqueue;
        private Mutex m_mut_trm;
        private Mutex m_mut_rcv;
        private AutoResetEvent m_are_trm;
        private AutoResetEvent m_are_rcv;
        private AutoResetEvent m_are_timeout_start;
        private AutoResetEvent m_are_timeout_stop;
        private AutoResetEvent m_are_timeout_is_stopped;
        #endregion
        #region PRIVATE_CONST_VALUES
        private const byte ACK                    = 0xC3;
        private const byte NACK                   = 0xCC;        
        private const byte ACK_MASK               = 0xC0;
        private const byte NOTHING                = 0x00;
        private const byte RESET                  = 0x33;        
        private const int  RBUFFER_SIZE           =  128;
        private const int  ACK_TIMEOUT            =  100;
        private const int  ACK_TIMEOUT_SHORT      =   30;
        private const int  RECEIVE_TIMEOUT        =   30;
        private const int  MAX_MESSAGE_LENGTH     = 0x16;
        private const int  MAX_TRANSMIT_MESSAGES  =    3;
        private const int  MESSAGE_LENGTH_POS     =    0;
        private const int  MESSAGE_SEQUCOUNT_POS  =    1;
        private const int  MESSAGE_CMD_POS        =    2;
        #endregion
    }
}
