/******************************************************************************
* BIM M 13x Serial Communication
*
* File: BIM_M_13x_Serial_LowLevel.c
*
* Date: 2009/03/16
*
* Version: 002
*
* Author: Schiekofer
******************************************************************************/
#include "BIM_M13x.h"
#include "BIM_M_13x_Serial_LowLevel.h"
#include "application.h"

#pragma dataseg = EXPRAM
__no_init BufferStruct SerialReceiveBuffer[SERIAL_NUM_OF_BUFFERS];
__no_init BufferStruct SerialTransmitBuffer[SERIAL_NUM_OF_BUFFERS];

__no_init SerialLowLevelData SLLD;
#pragma dataseg = default

const BYTE ResetTel[4] = {0x04, 0x00, RESET, ((0xA5 + 0x04 + RESET) & 0xFF)};
extern TIMER_TAB UserTimerTab;
extern void SLL_OnDataReceived(BYTE* pData, BYTE length);

// this is the interrupt handler, called for each received byte
void SerialReceiveInterrupt()
{
    // copy the received byte (and clear the receive buffer)
    BYTE rcvByte = RXB0;
    
    // check actual state
    switch (SLLD.rcvState)
    {
    case sll_rcv_idle:
        // first check if received byte was an ACK/NACK
        if ((rcvByte & ACK_MASK) == ACK_MASK)
        {
            // set the received ACK/NACK byte
            SLLD.rcvACK = rcvByte;
        }
        else
        {
            // start receiving of a message
            // set receive count
            SLLD.rcvCount = rcvByte - 1;
            // check receive count
            if (SLLD.rcvCount >= SERIAL_MAX_MESSAGE_LENGTH)
            {
                // limit receive count
                SLLD.rcvCount = SERIAL_MAX_MESSAGE_LENGTH - 1;
            }
            // indicate to start the receive timeout timer
            SLLD.timerctrl = sll_tmctrl_on;
            // check if buffer is free
            if (SerialReceiveBuffer[SLLD.rcvIndex].process == FALSE)
            {
                // go to state receiving
                SLLD.rcvState = sll_rcv_receiving;
                
                // set receive pointer to buffer
                SLLD.rcvPointer = SerialReceiveBuffer[SLLD.rcvIndex].Buffer;
                // move received byte into buffer
                *SLLD.rcvPointer = rcvByte;
                // increment receive pointer
                *SLLD.rcvPointer++;
            }
            else
            {
                // no free buffer -> wait end of message without receiving
                SLLD.rcvState = sll_rcv_waitend;
            }
        }
        break;
    case sll_rcv_receiving:
        // move received byte into buffer
        *SLLD.rcvPointer = rcvByte;
        // increment receive pointer
        *SLLD.rcvPointer++;
        // decrement receive counter
        SLLD.rcvCount--;
        // check counter == 0
        if (SLLD.rcvCount == 0)
        {
            // go to state idle
            SLLD.rcvState = sll_rcv_idle;
            // indicate to stop the receive timeout timer
            SLLD.timerctrl = sll_tmctrl_off;
            // indicate that the message in this buffer must be processed
            SerialReceiveBuffer[SLLD.rcvIndex].process = TRUE;
            // increment receive buffer index to next buffer
            SERIAL_INC_BUFFERINDEX(SLLD.rcvIndex);
        }        
        break;
    case sll_rcv_waitend:
        // just decrement receive counter
        SLLD.rcvCount--;
        // check counter == 0
        if (SLLD.rcvCount == 0)
        {
            // go to idle
            SLLD.rcvState = sll_rcv_idle;
        }
        break;
    default:
        // go to idle
        SLLD.rcvState = sll_rcv_idle;
        // reset pointer
        SLLD.rcvPointer = SerialReceiveBuffer[0].Buffer;
        break;
    }
}

// this is the interrupt handler, called for each transmitted byte
void SerialTransmitInterrupt()
{   
    // if transmit counter > 0
    if (SLLD.trmCount)
    {
        // move the next byte from the transmit pointer to the serial output register
        TXS0 = *SLLD.trmPointer;       
        // decrement the transmit counter
        SLLD.trmCount--;
        // increment the transmit pointer
        SLLD.trmPointer++;
    }
    // if there was an ACK/NACK moved into the serial output register
    else if (SLLD.trmACK != FALSE)
    {
        // reset the trmACK indicator
        SLLD.trmACK = FALSE;
    }
    else
    {
        // at this point transmitting of a message is over -> go to state sll_trm_transmit_finish
        SLLD.trmState = sll_trm_transmit_finish;
        // reset 'ACK/NACK received' indicator
        SLLD.rcvACK = FALSE;
    }        
}

void SerialLowLevel_Init()
{      
    // register serial transmit and receive interrupts
    U._IntRegister(SerialReceiveInterrupt, INTSR0_vect);
    U._IntRegister(SerialTransmitInterrupt, INTST0_vect);
    
    // configure serial port 0
    ASIM0 = 0xFD;           // even parity and one stop bit
    BRGC0 = 0xC8;           // 19,200 bits/s

    STIF0 = 0;              // clear interrupt ...
    SRIF0 = 0;              // ... flags
    STMK0 = 0;              // clear interrupt ...
    SERIAL_EN_RCVINT();     // ... masks
    
    P1_bit.no0 = 1;
    PM1_bit.no0 = 0;
    PM1_bit.no1 = 1;
}

void AddBufferChecksum(BYTE index)
{
    BYTE i;
    // start value of checksum is 0xA5
    BYTE checksum = 0xA5;
    for (i = 0; i < (SerialTransmitBuffer[index].count - 1); i++)
    {
        // add all bytes in the buffer
        checksum += SerialTransmitBuffer[index].Buffer[i];
    }
    // set calculated checksum into the last byte of tthe buffer
    SerialTransmitBuffer[index].Buffer[i] = checksum;
}

void Send(BYTE index)
{   
    // first check if a reset must be send
    if (SLLD.reset != sll_reset_idle)
    {
        // set transmit pointer to const reset message
        SLLD.trmPointer = ResetTel;
        // set transmit counter
        SLLD.trmCount = ResetTel[SERIAL_POS_LENGTH];
        // set reset confirm indicator
        SLLD.reset = sll_reset_conf;
    }
    else
    {
        // add transmit sequence counter
        SerialTransmitBuffer[index].Buffer[SERIAL_POS_SEQCOUNTER] = SLLD.trmSeqCounter;
        // add checksum
        AddBufferChecksum(index);
        // set transmit pointer
        SLLD.trmPointer = SerialTransmitBuffer[index].Buffer;
        // set transmit counter
        SLLD.trmCount = SerialTransmitBuffer[index].count;
    }
    // disable task switch here
    DIS_TASKSWITCH();
    // start transmitting
    SERIAL_START_SEND();              
    // go to state sll_trm_transmitting
    SLLD.trmState = sll_trm_transmitting;
    // enable task switch again
    EN_TASKSWITCH();
}

BOOL CheckBufferCheckSum(BYTE* ptr)
{
    BYTE length = *ptr;
    // checksum start with 0xA5
    BYTE checksum = 0xA5;
    for (BYTE i = 0; i < (length - 1); i++)
    {
        // add all byte 
        checksum += *ptr;
        ptr++;
    }
    // check if last byte in the message is equal to the calculated checksum
    return (checksum == *ptr) ? TRUE : FALSE;
}

// the following function is called cyclically
void SerialLowLevel_Task()
{ 
    // check transmit state
    switch (SLLD.trmState)
    {
    case sll_trm_idle:
        // is ACK/NACK send indicator set
        if (SLLD.trmACKByte)
        {
            // set indicator that now ACK/NACK is moved into the serial transmit register
            SLLD.trmACK = TRUE;
            // move ACK/NACK into the serial transmit register
            TXS0 = SLLD.trmACKByte;
            // reset indicator
            SLLD.trmACKByte = NULL;
        }
        else if (SLLD.trmACK != FALSE)
        {
            // do nothing while ACK/NACK is not transmitted
            break;
        }
        else if ((SerialTransmitBuffer[SLLD.trmIndex].process) || (SLLD.reset == sll_reset_req))
        {            
            // call 'Send' if there is a message to send or reset request is set
            Send(SLLD.trmIndex);
        }
        break;
    case sll_trm_transmitting:
        // do nothing while a message will be transmitted
        break;
    case sll_trm_transmit_finish:
        // start ACK/NACK timeout timer 
        U._TmStart(&UserTimerTab.Timers[TIMER_TIMEOUT], SERIAL_ACK_TIMEOUT);
        // go to sll_trm_waitforack
        SLLD.trmState = sll_trm_waitforack;
        break;
    case sll_trm_waitforack:
        // first check if there is an ACK/NACK to send
        if (SLLD.trmACKByte)
        {
            // set indicator that now ACK/NACK is moved into the serial transmit register
            SLLD.trmACK = TRUE;
            // move ACK/NACK into the serial transmit register
            TXS0 = SLLD.trmACKByte;
            // reset indicator
            SLLD.trmACKByte = NULL;
        }
        else if (SLLD.trmACK != FALSE)
        {
            // do nothing while ACK/NACK is not transmitted
            break;
        }
        // now check if an ACK/NACK was received
        else if (SLLD.rcvACK)
        {
            // stop ACK/NACK timeout timer
            U._TmStop(&UserTimerTab.Timers[TIMER_TIMEOUT]);
            if (SLLD.rcvACK == SERIAL_ACK)
            {
                // if a reset was send
                if (SLLD.reset == sll_reset_conf)
                {
                    // set reset indicator to idle
                    SLLD.reset = sll_reset_idle;
                    // reset receive sequence counter
                    SLLD.rcvSeqCounter = 0;
                }
                else
                {
                    // reset processing indicator for this message
                    SerialTransmitBuffer[SLLD.trmIndex].process = FALSE;
                    // set transmit index to next message buffer
                    SERIAL_INC_BUFFERINDEX(SLLD.trmIndex);
                }
                // increment transmit sequence counter
                SLLD.trmSeqCounter++;
                // go to sll_trm_idle
                SLLD.trmState = sll_trm_idle;                
            }
            else
            {
                // send message again
                Send(SLLD.trmIndex);
            }
        }
        else if (U._TmIsExpired(&UserTimerTab.Timers[TIMER_TIMEOUT]))
        {
            // ACK/NACK timeout timer is expired -> stop timer
            U._TmStop(&UserTimerTab.Timers[TIMER_TIMEOUT]);
            // send message again
            Send(SLLD.trmIndex);
        }
        break;
    default:
        // go to idle
        SLLD.trmState = sll_trm_idle;
        break;
    }    
    
    // check if serial receive timeout timer should be started
    if (SLLD.timerctrl == sll_tmctrl_on)
    {
        // set indicator to idle
        SLLD.timerctrl = sll_tmctrl_idle;
        // start timer
        U._TmStart(&UserTimerTab.Timers[TIMER_SERIAL], SERIAL_RCV_TIMEOUT);
    }
    // check if serial receive timeout timer should be stopped
    if (SLLD.timerctrl == sll_tmctrl_off)
    {
        // set indicator to idle
        SLLD.timerctrl = sll_tmctrl_idle;
        // stop timer
        U._TmStop(&UserTimerTab.Timers[TIMER_SERIAL]);
    }
    // check if serial receive timeout timer is expired
    if (U._TmIsExpired(&UserTimerTab.Timers[TIMER_SERIAL]))
    {
        // stop  timer
        U._TmStop(&UserTimerTab.Timers[TIMER_SERIAL]);
        // go to idle
        SLLD.rcvState = sll_rcv_idle;
        // indicate a NACK to send
        SLLD.trmACKByte = SERIAL_NACK;
    }
    // check if there is a message ready for processing
    if (SerialReceiveBuffer[SLLD.rcvIndexProcess].process)
    {
        // first check checksum
        if (CheckBufferCheckSum(SerialReceiveBuffer[SLLD.rcvIndexProcess].Buffer))
        {
            // check if message is a reset message
            if (SerialReceiveBuffer[SLLD.rcvIndexProcess].Buffer[SERIAL_POS_CMD] == RESET)
            {
                // indicate an ACK to send
                SLLD.trmACKByte = SERIAL_ACK;
                // increment receive sequence counter
                SLLD.rcvSeqCounter++;
                // reset transmit sequence counter
                SLLD.trmSeqCounter = 0;
                // if transmitter is waiting for ACK
                if (SLLD.trmState == sll_trm_waitforack)
                {
                    // stop and restart ACK/NACK timeout timer with a shorter timeout time
                    // to send the message faster again
                    U._TmStop(&UserTimerTab.Timers[TIMER_TIMEOUT]);
                    U._TmStart(&UserTimerTab.Timers[TIMER_TIMEOUT], SERIAL_ACK_TIMEOUT_SHORT);
                }
            }
            else if (SerialReceiveBuffer[SLLD.rcvIndexProcess].Buffer[SERIAL_POS_SEQCOUNTER] == SLLD.rcvSeqCounter)
            {
                // receive sequence counter is o.k. --> indicate an ACK to send
                SLLD.trmACKByte = SERIAL_ACK;
                // increment receive sequence counter
                SLLD.rcvSeqCounter++;
                // process data
                SLL_OnDataReceived(&SerialReceiveBuffer[SLLD.rcvIndexProcess].Buffer[2], SerialReceiveBuffer[SLLD.rcvIndexProcess].Buffer[0] - 3);
            }
            else
            {
                if (SerialReceiveBuffer[SLLD.rcvIndexProcess].Buffer[SERIAL_POS_SEQCOUNTER] == ((BYTE)(SLLD.rcvSeqCounter - 1)))
                {
                    // receive sequence counter is one lower than the expected one
                    // --> only send ACK (message was already processed)
                    SLLD.trmACKByte = SERIAL_ACK;
                }
                else
                {
                    // wrong receive sequence counter -> indicate a reset message to send
                    SLLD.reset = sll_reset_req;
                }            
            }   
        }
        else
        {
            // wrong checksum -> indicate a NACK to send
            SLLD.trmACKByte = SERIAL_NACK;
        }
               
        // reset process flag of the message
        SerialReceiveBuffer[SLLD.rcvIndexProcess].process = FALSE;
        // increment receive message index
        SERIAL_INC_BUFFERINDEX(SLLD.rcvIndexProcess);
    }
}

BOOL SLLSend(const BYTE* ptr, BYTE cmd, BYTE length)
{
    // check if message is too long
    if (length > SERIAL_MAX_MESSAGE_PAYLOAD)
    {
        return FALSE;
    }
    
    // search for next free transmit buffer
    BYTE i = SERIAL_NUM_OF_BUFFERS;
    BYTE j = SLLD.trmIndex;    
    while (i-- > 0)
    {
        if (SerialTransmitBuffer[j].process == FALSE)
        {
            break;
        }
        else
        {
            SERIAL_INC_BUFFERINDEX(j);
        }
    }
       
    if (SerialTransmitBuffer[j].process == FALSE)
    {        
        // free transmit buffer was found
        // set message length
        SerialTransmitBuffer[j].Buffer[SERIAL_POS_LENGTH] = length + SERIAL_LL_ADDED_BYTES;
        // set message count
        SerialTransmitBuffer[j].count = SerialTransmitBuffer[j].Buffer[SERIAL_POS_LENGTH];
        // set command byte
        SerialTransmitBuffer[j].Buffer[SERIAL_POS_CMD] = cmd;
        // copy message into transmit buffer
        for (BYTE i = 3; i < length + (SERIAL_LL_ADDED_BYTES - 1); i++)
        {
            SerialTransmitBuffer[j].Buffer[i] = *ptr;
            ptr++;
        }
        // indicate this message to transmit
        SerialTransmitBuffer[j].process = TRUE;
        return TRUE;
    }
    else
    {
        // no free transmit buffer was found -> return FALSE
        return FALSE;
    }
}
